import collections
import socket
import time
import logging
import six

from pymemcache.client.base import (
    Client,
    PooledClient,
    check_key_helper,
    normalize_server_spec,
)
from pymemcache.client.rendezvous import RendezvousHash
from pymemcache.exceptions import MemcacheError

logger = logging.getLogger(__name__)


class HashClient(object):
    """
    A client for communicating with a cluster of memcached servers
    """
    #: :class:`Client` class used to create new clients
    client_class = Client

    def __init__(
        self,
        servers,
        hasher=RendezvousHash,
        serde=None,
        serializer=None,
        deserializer=None,
        connect_timeout=None,
        timeout=None,
        no_delay=False,
        socket_module=socket,
        socket_keepalive=None,
        key_prefix=b'',
        max_pool_size=None,
        pool_idle_timeout=0,
        lock_generator=None,
        retry_attempts=2,
        retry_timeout=1,
        dead_timeout=60,
        use_pooling=False,
        ignore_exc=False,
        allow_unicode_keys=False,
        default_noreply=True,
        encoding='ascii',
        tls_context=None
    ):
        """
        Constructor.

        Args:
          servers: list() of tuple(hostname, port) or string containing a UNIX
                   socket path.
          hasher: optional class three functions ``get_node``, ``add_node``,
                  and ``remove_node``
                  defaults to Rendezvous (HRW) hash.

          use_pooling: use py:class:`.PooledClient` as the default underlying
                       class. ``max_pool_size`` and ``lock_generator`` can
                       be used with this. default: False

          retry_attempts: Amount of times a client should be tried before it
                          is marked dead and removed from the pool.
          retry_timeout (float): Time in seconds that should pass between retry
                                 attempts.
          dead_timeout (float): Time in seconds before attempting to add a node
                                back in the pool.
          encoding: optional str, controls data encoding (defaults to 'ascii').

        Further arguments are interpreted as for :py:class:`.Client`
        constructor.
        """
        self.clients = {}
        self.retry_attempts = retry_attempts
        self.retry_timeout = retry_timeout
        self.dead_timeout = dead_timeout
        self.use_pooling = use_pooling
        self.key_prefix = key_prefix
        self.ignore_exc = ignore_exc
        self.allow_unicode_keys = allow_unicode_keys
        self._failed_clients = {}
        self._dead_clients = {}
        self._last_dead_check_time = time.time()

        self.hasher = hasher()

        self.default_kwargs = {
            'connect_timeout': connect_timeout,
            'timeout': timeout,
            'no_delay': no_delay,
            'socket_module': socket_module,
            'socket_keepalive': socket_keepalive,
            'key_prefix': key_prefix,
            'serde': serde,
            'serializer': serializer,
            'deserializer': deserializer,
            'allow_unicode_keys': allow_unicode_keys,
            'default_noreply': default_noreply,
            'encoding': encoding,
            'tls_context': tls_context,
        }

        if use_pooling is True:
            self.default_kwargs.update({
                'max_pool_size': max_pool_size,
                'pool_idle_timeout': pool_idle_timeout,
                'lock_generator': lock_generator
            })

        for server in servers:
            self.add_server(normalize_server_spec(server))
        self.encoding = encoding
        self.tls_context = tls_context

    def _make_client_key(self, server):
        if isinstance(server, (list, tuple)) and len(server) == 2:
            return '%s:%s' % server
        return server

    def add_server(self, server, port=None):
        # To maintain backward compatibility, if a port is provided, assume
        # that server wasn't provided as a (host, port) tuple.
        if port is not None:
            if not isinstance(server, six.string_types):
                raise TypeError('Server must be a string when passing port.')
            server = (server, port)

        _class = PooledClient if self.use_pooling else self.client_class
        client = _class(server, **self.default_kwargs)
        if self.use_pooling:
            client.client_class = self.client_class

        key = self._make_client_key(server)
        self.clients[key] = client
        self.hasher.add_node(key)

    def remove_server(self, server, port=None):
        # To maintain backward compatibility, if a port is provided, assume
        # that server wasn't provided as a (host, port) tuple.
        if port is not None:
            if not isinstance(server, six.string_types):
                raise TypeError('Server must be a string when passing port.')
            server = (server, port)

        key = self._make_client_key(server)
        dead_time = time.time()
        self._failed_clients.pop(server)
        self._dead_clients[server] = dead_time
        self.hasher.remove_node(key)

    def _retry_dead(self):
        current_time = time.time()
        ldc = self._last_dead_check_time
        # We have reached the retry timeout
        if current_time - ldc > self.dead_timeout:
            candidates = []
            for server, dead_time in self._dead_clients.items():
                if current_time - dead_time > self.dead_timeout:
                    candidates.append(server)
            for server in candidates:
                logger.debug(
                    'bringing server back into rotation %s',
                    server
                )
                self.add_server(server)
                del self._dead_clients[server]
            self._last_dead_check_time = current_time

    def _get_client(self, key):
        check_key_helper(key, self.allow_unicode_keys, self.key_prefix)
        if self._dead_clients:
            self._retry_dead()

        server = self.hasher.get_node(key)
        # We've ran out of servers to try
        if server is None:
            if self.ignore_exc is True:
                return
            raise MemcacheError('All servers seem to be down right now')

        return self.clients[server]

    def _safely_run_func(self, client, func, default_val, *args, **kwargs):
        try:
            if client.server in self._failed_clients:
                # This server is currently failing, lets check if it is in
                # retry or marked as dead
                failed_metadata = self._failed_clients[client.server]

                # we haven't tried our max amount yet, if it has been enough
                # time lets just retry using it
                if failed_metadata['attempts'] < self.retry_attempts:
                    failed_time = failed_metadata['failed_time']
                    if time.time() - failed_time > self.retry_timeout:
                        logger.debug(
                            'retrying failed server: %s', client.server
                        )
                        result = func(*args, **kwargs)
                        # we were successful, lets remove it from the failed
                        # clients
                        self._failed_clients.pop(client.server)
                        return result
                    return default_val
                else:
                    # We've reached our max retry attempts, we need to mark
                    # the sever as dead
                    logger.debug('marking server as dead: %s', client.server)
                    self.remove_server(client.server)

            result = func(*args, **kwargs)
            return result

        # Connecting to the server fail, we should enter
        # retry mode
        except socket.error:
            self._mark_failed_server(client.server)

            # if we haven't enabled ignore_exc, don't move on gracefully, just
            # raise the exception
            if not self.ignore_exc:
                raise

            return default_val
        except Exception:
            # any exceptions that aren't socket.error we need to handle
            # gracefully as well
            if not self.ignore_exc:
                raise

            return default_val

    def _safely_run_set_many(self, client, values, *args, **kwargs):
        failed = []
        succeeded = []
        try:
            if client.server in self._failed_clients:
                # This server is currently failing, lets check if it is in
                # retry or marked as dead
                failed_metadata = self._failed_clients[client.server]

                # we haven't tried our max amount yet, if it has been enough
                # time lets just retry using it
                if failed_metadata['attempts'] < self.retry_attempts:
                    failed_time = failed_metadata['failed_time']
                    if time.time() - failed_time > self.retry_timeout:
                        logger.debug(
                            'retrying failed server: %s', client.server
                        )
                        succeeded, failed, err = self._set_many(
                            client, values, *args, **kwargs)
                        if err is not None:
                            raise err
                        # we were successful, lets remove it from the failed
                        # clients
                        self._failed_clients.pop(client.server)
                        return failed
                    return values.keys()
                else:
                    # We've reached our max retry attempts, we need to mark
                    # the sever as dead
                    logger.debug('marking server as dead: %s', client.server)
                    self.remove_server(client.server)

            succeeded, failed, err = self._set_many(
                client, values, *args, **kwargs
            )
            if err is not None:
                raise err

            return failed

        # Connecting to the server fail, we should enter
        # retry mode
        except socket.error:
            self._mark_failed_server(client.server)

            # if we haven't enabled ignore_exc, don't move on gracefully, just
            # raise the exception
            if not self.ignore_exc:
                raise

            return list(set(values.keys()) - set(succeeded))
        except Exception:
            # any exceptions that aren't socket.error we need to handle
            # gracefully as well
            if not self.ignore_exc:
                raise

            return list(set(values.keys()) - set(succeeded))

    def _mark_failed_server(self, server):
        # This client has never failed, lets mark it for failure
        if (
                server not in self._failed_clients and
                self.retry_attempts > 0
        ):
            self._failed_clients[server] = {
                'failed_time': time.time(),
                'attempts': 0,
            }
        # We aren't allowing any retries, we should mark the server as
        # dead immediately
        elif (
            server not in self._failed_clients and
            self.retry_attempts <= 0
        ):
            self._failed_clients[server] = {
                'failed_time': time.time(),
                'attempts': 0,
            }
            logger.debug("marking server as dead %s", server)
            self.remove_server(server)
        # This client has failed previously, we need to update the metadata
        # to reflect that we have attempted it again
        else:
            failed_metadata = self._failed_clients[server]
            failed_metadata['attempts'] += 1
            failed_metadata['failed_time'] = time.time()
            self._failed_clients[server] = failed_metadata

    def _run_cmd(self, cmd, key, default_val, *args, **kwargs):
        client = self._get_client(key)

        if client is None:
            return default_val

        func = getattr(client, cmd)
        args = list(args)
        args.insert(0, key)
        return self._safely_run_func(
            client, func, default_val, *args, **kwargs
        )

    def _set_many(self, client, values, *args, **kwargs):
        failed = []
        succeeded = []

        try:
            failed = client.set_many(values, *args, **kwargs)
        except Exception as e:
            if not self.ignore_exc:
                return succeeded, failed, e

        succeeded = [key for key in six.iterkeys(values) if key not in failed]
        return succeeded, failed, None

    def close(self):
        for client in self.clients.values():
            self._safely_run_func(client, client.close, False)

    disconnect_all = close

    def set(self, key, *args, **kwargs):
        return self._run_cmd('set', key, False, *args, **kwargs)

    def get(self, key, *args, **kwargs):
        return self._run_cmd('get', key, None, *args, **kwargs)

    def incr(self, key, *args, **kwargs):
        return self._run_cmd('incr', key, False, *args, **kwargs)

    def decr(self, key, *args, **kwargs):
        return self._run_cmd('decr', key, False, *args, **kwargs)

    def set_many(self, values, *args, **kwargs):
        client_batches = collections.defaultdict(dict)
        failed = []

        for key, value in six.iteritems(values):
            client = self._get_client(key)

            if client is None:
                failed.append(key)
                continue

            client_batches[client.server][key] = value

        for server, values in client_batches.items():
            client = self.clients[self._make_client_key(server)]
            failed += self._safely_run_set_many(
                client, values, *args, **kwargs
            )

        return failed

    set_multi = set_many

    def get_many(self, keys, gets=False, *args, **kwargs):
        client_batches = collections.defaultdict(list)
        end = {}

        for key in keys:
            client = self._get_client(key)

            if client is None:
                continue

            client_batches[client.server].append(key)

        for server, keys in client_batches.items():
            client = self.clients[self._make_client_key(server)]
            new_args = list(args)
            new_args.insert(0, keys)

            if gets:
                get_func = client.gets_many
            else:
                get_func = client.get_many

            result = self._safely_run_func(
                client,
                get_func, {}, *new_args, **kwargs
            )
            end.update(result)

        return end

    get_multi = get_many

    def gets(self, key, *args, **kwargs):
        return self._run_cmd('gets', key, None, *args, **kwargs)

    def gets_many(self, keys, *args, **kwargs):
        return self.get_many(keys, gets=True, *args, **kwargs)

    gets_multi = gets_many

    def add(self, key, *args, **kwargs):
        return self._run_cmd('add', key, False, *args, **kwargs)

    def prepend(self, key, *args, **kwargs):
        return self._run_cmd('prepend', key, False, *args, **kwargs)

    def append(self, key, *args, **kwargs):
        return self._run_cmd('append', key, False, *args, **kwargs)

    def delete(self, key, *args, **kwargs):
        return self._run_cmd('delete', key, False, *args, **kwargs)

    def delete_many(self, keys, *args, **kwargs):
        for key in keys:
            self._run_cmd('delete', key, False, *args, **kwargs)
        return True

    delete_multi = delete_many

    def cas(self, key, *args, **kwargs):
        return self._run_cmd('cas', key, False, *args, **kwargs)

    def replace(self, key, *args, **kwargs):
        return self._run_cmd('replace', key, False, *args, **kwargs)

    def touch(self, key, *args, **kwargs):
        return self._run_cmd('touch', key, False, *args, **kwargs)

    def flush_all(self):
        for client in self.clients.values():
            self._safely_run_func(client, client.flush_all, False)

    def quit(self):
        for client in self.clients.values():
            self._safely_run_func(client, client.quit, False)



import six
import pytest

from pymemcache.test.utils import MockMemcacheClient


@pytest.mark.unit()
def test_get_set():
    client = MockMemcacheClient()
    assert client.get(b"hello") is None

    client.set(b"hello", 12)
    assert client.get(b"hello") == 12


@pytest.mark.unit()
def test_get_set_unicide_key():
    client = MockMemcacheClient()
    assert client.get(u"hello") is None

    client.set(b"hello", 12)
    assert client.get(u"hello") == 12


@pytest.mark.unit()
def test_get_set_non_ascii_value():
    client = MockMemcacheClient()
    assert client.get(b"hello") is None

    # This is the value of msgpack.packb('non_ascii')
    non_ascii_str = b'\xa9non_ascii'
    client.set(b"hello", non_ascii_str)
    assert client.get(b"hello") == non_ascii_str


@pytest.mark.unit()
def test_get_many_set_many():
    client = MockMemcacheClient()
    client.set(b"h", 1)

    result = client.get_many([b"h", b"e", b"l", b"o"])
    assert result == {b"h": 1}

    # Convert keys into bytes
    d = dict((k.encode('ascii'), v)
             for k, v in six.iteritems(dict(h=1, e=2, z=3)))
    client.set_many(d)
    assert client.get_many([b"h", b"e", b"z", b"o"]) == d


@pytest.mark.unit()
def test_get_many_set_many_non_ascii_values():
    client = MockMemcacheClient()

    # These are the values of calling msgpack.packb() on '1', '2', and '3'
    non_ascii_1 = b'\xa11'
    non_ascii_2 = b'\xa12'
    non_ascii_3 = b'\xa13'
    client.set(b"h", non_ascii_1)

    result = client.get_many([b"h", b"e", b"l", b"o"])
    assert result == {b"h": non_ascii_1}

    # Convert keys into bytes
    d = dict((k.encode('ascii'), v)
             for k, v in six.iteritems(
                dict(h=non_ascii_1, e=non_ascii_2, z=non_ascii_3)
             ))
    client.set_many(d)
    assert client.get_many([b"h", b"e", b"z", b"o"]) == d


@pytest.mark.unit()
def test_add():
    client = MockMemcacheClient()

    client.add(b"k", 2)
    assert client.get(b"k") == 2

    client.add(b"k", 25)
    assert client.get(b"k") == 2


@pytest.mark.unit()
def test_delete():
    client = MockMemcacheClient()

    client.add(b"k", 2)
    assert client.get(b"k") == 2

    client.delete(b"k")
    assert client.get(b"k") is None


@pytest.mark.unit()
def test_incr_decr():
    client = MockMemcacheClient()

    client.add(b"k", 2)

    client.incr(b"k", 4)
    assert client.get(b"k") == 6

    client.decr(b"k", 2)
    assert client.get(b"k") == 4


@pytest.mark.unit()
def test_prepand_append():
    client = MockMemcacheClient()

    client.set(b"k", '1')
    client.append(b"k", 'a')
    client.prepend(b"k", 'p')
    assert client.get(b"k") == b'p1a'



# API Backwards compatibility

from pymemcache.client.base import Client  # noqa
from pymemcache.client.base import PooledClient  # noqa
from pymemcache.client.hash import HashClient  # noqa
from pymemcache.client.retrying import RetryingClient  # noqa

from pymemcache.exceptions import MemcacheError  # noqa
from pymemcache.exceptions import MemcacheClientError  # noqa
from pymemcache.exceptions import MemcacheUnknownCommandError  # noqa
from pymemcache.exceptions import MemcacheIllegalInputError  # noqa
from pymemcache.exceptions import MemcacheServerError  # noqa
from pymemcache.exceptions import MemcacheUnknownError  # noqa
from pymemcache.exceptions import MemcacheUnexpectedCloseError  # noqa



import os.path
import pytest
import socket
import ssl


def pytest_addoption(parser):
    parser.addoption('--server', action='store', default='localhost',
                     help='memcached server')

    parser.addoption('--port', action='store', default='11211',
                     help='memcached server port')

    parser.addoption('--tls-server', action='store', default='localhost',
                     help='TLS memcached server')

    parser.addoption('--tls-port', action='store', default='11212',
                     help='TLS memcached server port')

    parser.addoption('--size', action='store', default=1024,
                     help='size of data in benchmarks')

    parser.addoption('--count', action='store', default=10000,
                     help='number of iterations to run each benchmark')

    parser.addoption('--keys', action='store', default=20,
                     help='number of keys to use for multi benchmarks')


@pytest.fixture(scope='session')
def host(request):
    return request.config.option.server


@pytest.fixture(scope='session')
def port(request):
    return int(request.config.option.port)


@pytest.fixture(scope='session')
def tls_host(request):
    return request.config.option.tls_server


@pytest.fixture(scope='session')
def tls_port(request):
    return int(request.config.option.tls_port)


@pytest.fixture(scope='session')
def size(request):
    return int(request.config.option.size)


@pytest.fixture(scope='session')
def count(request):
    return int(request.config.option.count)


@pytest.fixture(scope='session')
def keys(request):
    return int(request.config.option.keys)


@pytest.fixture(scope='session')
def pairs(size, keys):
    return {'pymemcache_test:%d' % i: 'X' * size for i in range(keys)}


@pytest.fixture(scope='session')
def tls_context():
    return ssl.create_default_context(
        cafile=os.path.join(os.path.dirname(__file__), "certs/ca-root.crt")
    )


def pytest_generate_tests(metafunc):
    if 'socket_module' in metafunc.fixturenames:
        socket_modules = [socket]
        try:
            from gevent import socket as gevent_socket
        except ImportError:
            print("Skipping gevent (not installed)")
        else:
            socket_modules.append(gevent_socket)

        metafunc.parametrize("socket_module", socket_modules)

    if 'client_class' in metafunc.fixturenames:
        from pymemcache.client.base import PooledClient, Client
        from pymemcache.client.hash import HashClient

        class HashClientSingle(HashClient):
            def __init__(self, server, *args, **kwargs):
                super(HashClientSingle, self).__init__(
                    [server], *args, **kwargs
                )

        metafunc.parametrize(
            "client_class", [Client, PooledClient, HashClientSingle]
        )



# Copyright 2012 Pinterest.com
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from builtins import bytes as newbytes

import collections
import errno
import functools
import json
import os
import mock
import platform
import re
import socket
import sys
import unittest

import pytest

from pymemcache.client.base import (
    PooledClient,
    Client,
    normalize_server_spec,
    KeepaliveOpts,
    check_key_helper
)
from pymemcache.exceptions import (
    MemcacheClientError,
    MemcacheServerError,
    MemcacheUnexpectedCloseError,
    MemcacheUnknownCommandError,
    MemcacheUnknownError,
    MemcacheIllegalInputError
)

from pymemcache import pool
from pymemcache.test.utils import MockMemcacheClient


# TODO: Use ipaddress module when dropping support for Python < 3.3
def is_ipv6(address):
    return re.match(r'^[0-9a-f:]+$', address)


@pytest.mark.parametrize(
    'key,allow_unicode_keys,key_prefix,ex_exception,ex_excinfo,ignore_py27',
    [
        (u'b'*251, True, b'',
         MemcacheIllegalInputError, 'Key is too long', False),
        (u'foo bar', True, b'',
         MemcacheIllegalInputError, 'Key contains whitespace', False),
        (u'\00', True, b'',
         MemcacheIllegalInputError, 'Key contains null', False),
        (None, True, b'', TypeError, None, False),
        # The following test won't fail with a TypeError with python 2.7
        (b"", False, '', TypeError, None, True),
    ])
@pytest.mark.unit()
def test_check_key_helper_failing_conditions(key, allow_unicode_keys,
                                             key_prefix, ex_exception,
                                             ex_excinfo, ignore_py27):

    if ignore_py27 and sys.version_info < (3, 0, 0):
        pytest.skip("skipping for Python 2.7")
    with pytest.raises(ex_exception) as excinfo:
        check_key_helper(key, allow_unicode_keys, key_prefix)

    # Allow to ignore the excinfo value. Different implementation (2.7,
    # pypy, 3.x) comes with various error messages for the same thing.
    # We just check the kind of the exception.
    if ex_excinfo:
        assert ex_excinfo in str(excinfo.value)


@pytest.mark.unit()
def test_check_key_helper():
    assert check_key_helper(b"key", True, b'') == b"key"
    assert check_key_helper("key", True) == b"key"
    assert isinstance(check_key_helper(b"key", True), bytes)
    assert check_key_helper("", True) == b""


class MockSocket(object):
    def __init__(self, recv_bufs, connect_failure=None, close_failure=None):
        self.recv_bufs = collections.deque(recv_bufs)
        self.send_bufs = []
        self.closed = False
        self.timeouts = []
        self.connect_failure = connect_failure
        self.close_failure = close_failure
        self.connections = []
        self.socket_options = []

    @property
    def family(self):
        any_ipv6 = any(is_ipv6(c[0]) for c in self.connections)
        return socket.AF_INET6 if any_ipv6 else socket.AF_INET

    def sendall(self, value):
        self.send_bufs.append(value)

    def close(self):
        if isinstance(self.close_failure, Exception):
            raise self.close_failure
        self.closed = True

    def recv(self, size):
        value = self.recv_bufs.popleft()
        if isinstance(value, Exception):
            raise value
        return value

    def settimeout(self, timeout):
        self.timeouts.append(timeout)

    def connect(self, server):
        if isinstance(self.connect_failure, Exception):
            raise self.connect_failure
        self.connections.append(server)

    def setsockopt(self, level, option, value):
        self.socket_options.append((level, option, value))


class MockUnixSocketServer(object):
    def __init__(self, socket_path):
        if os.path.exists(socket_path):
            os.remove(socket_path)
        self.socket_path = socket_path
        self.socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

    def __enter__(self):
        self.socket.bind(self.socket_path)
        self.socket.listen(1)
        return self.socket

    def __exit__(self, *args):
        self.socket.close()
        os.remove(self.socket_path)


class MockSocketModule(object):
    def __init__(self, connect_failure=None, close_failure=None):
        self.connect_failure = connect_failure
        self.close_failure = close_failure
        self.sockets = []

    def socket(self, family, type, proto=0, fileno=None):
        socket = MockSocket(
            [],
            connect_failure=self.connect_failure,
            close_failure=self.close_failure)
        self.sockets.append(socket)
        return socket

    def getaddrinfo(self, host, port, family=0, type=0, proto=0, flags=0):
        family = family or (
            socket.AF_INET6 if is_ipv6(host) else socket.AF_INET
        )
        type = type or socket.SOCK_STREAM
        proto = proto or socket.IPPROTO_TCP
        sockaddr = (
            ('::1', 11211, 0, 0)
            if family == socket.AF_INET6
            else ('127.0.0.1', 11211)
        )
        return [(family, type, proto, '', sockaddr)]

    def __getattr__(self, name):
        return getattr(socket, name)


class CustomizedClient(Client):

    def _extract_value(self, expect_cas, line, buf, remapped_keys,
                       prefixed_keys):
        return b'key', b'value', b'END\r\n'


@pytest.mark.unit()
class ClientTestMixin(object):
    def make_client(self, mock_socket_values, **kwargs):
        client = Client(None, **kwargs)
        # mock out client._connect() rather than hard-settting client.sock to
        # ensure methods are checking whether self.sock is None before
        # attempting to use it
        sock = MockSocket(list(mock_socket_values))
        client._connect = mock.Mock(side_effect=functools.partial(
            setattr, client, "sock", sock))
        return client

    def make_customized_client(self, mock_socket_values, **kwargs):
        client = CustomizedClient(None, **kwargs)
        # mock out client._connect() rather than hard-settting client.sock to
        # ensure methods are checking whether self.sock is None before
        # attempting to use it
        sock = MockSocket(list(mock_socket_values))
        client._connect = mock.Mock(side_effect=functools.partial(
            setattr, client, "sock", sock))
        return client

    def test_set_success(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.set(b'key', b'value', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set(b'key', b'value', noreply=False)
        assert result is True

        # unit test for set operation with parameter flags
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set(b'key', b'value', noreply=False, flags=0x00000030)
        assert result is True

    def test_set_future(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.set(newbytes(b'key'), newbytes(b'value'), noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set(newbytes(b'key'), newbytes(b'value'), noreply=False)
        assert result is True

        # unit test for set operation with parameter flags
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set(newbytes(b'key'), newbytes(b'value'), noreply=False,
                            flags=0x00000030)
        assert result is True

    def test_set_unicode_key(self):
        client = self.make_client([b''])

        def _set():
            client.set(u'\u0FFF', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_unicode_key_ok(self):
        client = self.make_client([b'STORED\r\n'], allow_unicode_keys=True)

        result = client.set(u'\u0FFF', b'value', noreply=False)
        assert result is True

    def test_set_unicode_key_ok_snowman(self):
        client = self.make_client([b'STORED\r\n'], allow_unicode_keys=True)

        result = client.set('my☃', b'value', noreply=False)
        assert result is True

    def test_set_unicode_char_in_middle_of_key(self):
        client = self.make_client([b'STORED\r\n'])

        def _set():
            client.set('helloworld_\xb1901520_%c3', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_unicode_char_in_middle_of_key_snowman(self):
        client = self.make_client([b'STORED\r\n'])

        def _set():
            client.set('my☃', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_unicode_value(self):
        client = self.make_client([b''])

        def _set():
            client.set(b'key', u'\u0FFF', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_unicode_char_in_middle_of_key_ok(self):
        client = self.make_client([b'STORED\r\n'], allow_unicode_keys=True)

        result = client.set('helloworld_\xb1901520_%c3', b'value',
                            noreply=False)
        assert result is True

    def test_set_noreply(self):
        client = self.make_client([])
        result = client.set(b'key', b'value', noreply=True)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([], encoding='utf-8')
        result = client.set(b'key', b'value', noreply=True)
        assert result is True

    def test_set_many_success(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.set_many({b'key': b'value'}, noreply=False)
        assert result == []

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set_many({b'key': b'value'}, noreply=False)
        assert result == []

    def test_set_multi_success(self):
        # Should just map to set_many
        client = self.make_client([b'STORED\r\n'])
        result = client.set_multi({b'key': b'value'}, noreply=False)
        assert result == []

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set_multi({b'key': b'value'}, noreply=False)
        assert result == []

    def test_add_stored(self):
        client = self.make_client([b'STORED\r', b'\n'])
        result = client.add(b'key', b'value', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r', b'\n'], encoding='utf-8')
        result = client.add(b'key', b'value', noreply=False)
        assert result is True

    def test_add_not_stored(self):
        client = self.make_client([b'STORED\r', b'\n',
                                   b'NOT_', b'STOR', b'ED', b'\r\n'])
        client.add(b'key', b'value', noreply=False)
        result = client.add(b'key', b'value', noreply=False)
        assert result is False

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r', b'\n',
                                   b'NOT_', b'STOR', b'ED', b'\r\n'],
                                  encoding='utf-8')
        client.add(b'key', b'value', noreply=False)
        result = client.add(b'key', b'value', noreply=False)
        assert result is False

    def test_get_not_found(self):
        client = self.make_client([b'END\r\n'])
        result = client.get(b'key')
        assert result is None

        # Unit test for customized client (override _extract_value)
        client = self.make_customized_client([b'END\r\n'])
        result = client.get(b'key')
        assert result is None

    def test_space_key(self):
        client = self.make_client([b''])
        with pytest.raises(MemcacheIllegalInputError):
            client.get(b'space key')

        with pytest.raises(MemcacheIllegalInputError):
            client.set(b'space key', b'value')

    def test_get_not_found_default(self):
        client = self.make_client([b'END\r\n'])
        result = client.get(b'key', default='foobar')
        assert result == 'foobar'

        # Unit test for customized client (override _extract_value)
        client = self.make_customized_client([b'END\r\n'])
        result = client.get(b'key', default='foobar')
        assert result == 'foobar'

    def test_get_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'VALUE key 0 5\r\nvalue\r\nEND\r\n',
        ])
        client.set(b'key', b'value', noreply=False)
        result = client.get(b'key')
        assert result == b'value'

        # Unit test for customized client (override _extract_value)
        client = self.make_customized_client([
            b'STORED\r\n',
            b'VALUE key 0 5\r\nvalue\r\nEND\r\n',
        ])
        client.set(b'key', b'value', noreply=False)
        result = client.get(b'key')
        assert result == b'value'

    def test_get_many_none_found(self):
        client = self.make_client([b'END\r\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {}

    def test_get_multi_none_found(self):
        client = self.make_client([b'END\r\n'])
        result = client.get_multi([b'key1', b'key2'])
        assert result == {}

    def test_get_many_some_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'VALUE key1 0 6\r\nvalue1\r\nEND\r\n',
        ])
        client.set(b'key1', b'value1', noreply=False)
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1'}

    def test_get_many_all_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'STORED\r\n',
            b'VALUE key1 0 6\r\nvalue1\r\n',
            b'VALUE key2 0 6\r\nvalue2\r\nEND\r\n',
        ])
        client.set(b'key1', b'value1', noreply=False)
        client.set(b'key2', b'value2', noreply=False)
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

    def test_get_unicode_key(self):
        client = self.make_client([b''])

        def _get():
            client.get(u'\u0FFF')

        with pytest.raises(MemcacheIllegalInputError):
            _get()

    def test_delete_not_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.delete(b'key', noreply=False)
        assert result is False

    def test_delete_found(self):
        client = self.make_client([b'STORED\r', b'\n', b'DELETED\r\n'])
        client.add(b'key', b'value', noreply=False)
        result = client.delete(b'key', noreply=False)
        assert result is True

    def test_delete_noreply(self):
        client = self.make_client([])
        result = client.delete(b'key', noreply=True)
        assert result is True

    def test_delete_many_no_keys(self):
        client = self.make_client([])
        result = client.delete_many([], noreply=False)
        assert result is True

    def test_delete_many_none_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.delete_many([b'key'], noreply=False)
        assert result is True

    def test_delete_many_found(self):
        client = self.make_client([b'STORED\r', b'\n', b'DELETED\r\n'])
        client.add(b'key', b'value', noreply=False)
        result = client.delete_many([b'key'], noreply=False)
        assert result is True

    def test_delete_many_some_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'DELETED\r\n',
            b'NOT_FOUND\r\n'
        ])
        client.add(b'key', b'value', noreply=False)
        result = client.delete_many([b'key', b'key2'], noreply=False)
        assert result is True

    def test_delete_multi_some_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'DELETED\r\n',
            b'NOT_FOUND\r\n'
        ])
        client.add(b'key', b'value', noreply=False)
        result = client.delete_multi([b'key', b'key2'], noreply=False)
        assert result is True

    def test_incr_not_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.incr(b'key', 1, noreply=False)
        assert result is None

    def test_incr_found(self):
        client = self.make_client([b'STORED\r\n', b'1\r\n'])
        client.set(b'key', 0, noreply=False)
        result = client.incr(b'key', 1, noreply=False)
        assert result == 1

    def test_incr_noreply(self):
        client = self.make_client([b'STORED\r\n'])
        client.set(b'key', 0, noreply=False)

        client = self.make_client([])
        result = client.incr(b'key', 1, noreply=True)
        assert result is None

    def test_decr_not_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.decr(b'key', 1, noreply=False)
        assert result is None

    def test_decr_found(self):
        client = self.make_client([b'STORED\r\n', b'1\r\n'])
        client.set(b'key', 2, noreply=False)
        result = client.decr(b'key', 1, noreply=False)
        assert result == 1


class TestClient(ClientTestMixin, unittest.TestCase):

    Client = Client

    def test_append_stored(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.append(b'key', b'value', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.append(b'key', b'value', noreply=False)
        assert result is True

    def test_prepend_stored(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.prepend(b'key', b'value', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.prepend(b'key', b'value', noreply=False)
        assert result is True

    def test_cas_malformed(self):
        client = self.make_client([b'STORED\r\n'])
        with pytest.raises(MemcacheIllegalInputError):
            client.cas(b'key', b'value', None, noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            client.cas(b'key', b'value', 'nonintegerstring', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            # even a space makes it a noninteger string
            client.cas(b'key', b'value', '123 ', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            # non-ASCII digit
            client.cas(b'key', b'value', u'⁰', noreply=False)

    def test_cas_stored(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is True

    def test_cas_exists(self):
        client = self.make_client([b'EXISTS\r\n'])
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is False

        # unit test for encoding passed in __init__()
        client = self.make_client([b'EXISTS\r\n'], encoding='utf-8')
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is False

    def test_cas_not_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is None

        # unit test for encoding passed in __init__()
        client = self.make_client([b'NOT_FOUND\r\n'], encoding='utf-8')
        result = client.cas(b'key', b'value', b'123', noreply=False)
        assert result is None

    def test_cr_nl_boundaries(self):
        client = self.make_client([b'VALUE key1 0 6\r',
                                   b'\nvalue1\r\n'
                                   b'VALUE key2 0 6\r\n',
                                   b'value2\r\n'
                                   b'END\r\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

        client = self.make_client([b'VALUE key1 0 6\r\n',
                                   b'value1\r',
                                   b'\nVALUE key2 0 6\r\n',
                                   b'value2\r\n',
                                   b'END\r\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

        client = self.make_client([b'VALUE key1 0 6\r\n',
                                   b'value1\r\n',
                                   b'VALUE key2 0 6\r',
                                   b'\nvalue2\r\n',
                                   b'END\r\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

        client = self.make_client([b'VALUE key1 0 6\r\n',
                                   b'value1\r\n',
                                   b'VALUE key2 0 6\r\n',
                                   b'value2\r',
                                   b'\nEND\r\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

        client = self.make_client([b'VALUE key1 0 6\r\n',
                                   b'value1\r\n',
                                   b'VALUE key2 0 6\r\n',
                                   b'value2\r\n',
                                   b'END\r',
                                   b'\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

        client = self.make_client([b'VALUE key1 0 6\r',
                                   b'\nvalue1\r',
                                   b'\nVALUE key2 0 6\r',
                                   b'\nvalue2\r',
                                   b'\nEND\r',
                                   b'\n'])
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

    def test_delete_exception(self):
        client = self.make_client([Exception('fail')])

        def _delete():
            client.delete(b'key', noreply=False)

        with pytest.raises(Exception):
            _delete()

        assert client.sock is None

    def test_flush_all(self):
        client = self.make_client([b'OK\r\n'])
        result = client.flush_all(noreply=False)
        assert result is True

    def test_incr_exception(self):
        client = self.make_client([Exception('fail')])

        def _incr():
            client.incr(b'key', 1)

        with pytest.raises(Exception):
            _incr()

        assert client.sock is None

    def test_get_error(self):
        client = self.make_client([b'ERROR\r\n'])

        def _get():
            client.get(b'key')

        with pytest.raises(MemcacheUnknownCommandError):
            _get()

    def test_get_recv_chunks(self):
        client = self.make_client([b'VALUE key', b' 0 5\r', b'\nvalue',
                                   b'\r\n', b'END', b'\r', b'\n'])
        result = client.get(b'key')
        assert result == b'value'

    def test_get_unknown_error(self):
        client = self.make_client([b'foobarbaz\r\n'])

        def _get():
            client.get(b'key')

        with pytest.raises(MemcacheUnknownError):
            _get()

    def test_gets_not_found(self):
        client = self.make_client([b'END\r\n'])
        result = client.gets(b'key')
        assert result == (None, None)

    def test_gets_not_found_defaults(self):
        client = self.make_client([b'END\r\n'])
        result = client.gets(b'key', default='foo', cas_default='bar')
        assert result == ('foo', 'bar')

    def test_gets_found(self):
        client = self.make_client([b'VALUE key 0 5 10\r\nvalue\r\nEND\r\n'])
        result = client.gets(b'key')
        assert result == (b'value', b'10')

    def test_gets_many_none_found(self):
        client = self.make_client([b'END\r\n'])
        result = client.gets_many([b'key1', b'key2'])
        assert result == {}

    def test_gets_many_some_found(self):
        client = self.make_client([b'VALUE key1 0 6 11\r\nvalue1\r\nEND\r\n'])
        result = client.gets_many([b'key1', b'key2'])
        assert result == {b'key1': (b'value1', b'11')}

    def test_touch_not_found(self):
        client = self.make_client([b'NOT_FOUND\r\n'])
        result = client.touch(b'key', noreply=False)
        assert result is False

    def test_touch_found(self):
        client = self.make_client([b'TOUCHED\r\n'])
        result = client.touch(b'key', noreply=False)
        assert result is True

    def test_quit(self):
        client = self.make_client([])
        result = client.quit()
        assert result is None
        assert client.sock is None

    def test_shutdown(self):
        client = self.make_client([MemcacheUnexpectedCloseError('shutdown')])
        result = client.shutdown()
        assert result is None

    def test_shutdown_disabled(self):
        def _shutdown():
            client = self.make_client([b'ERROR: shutdown not enabled\r\n'])
            client.shutdown()

        with pytest.raises(MemcacheUnknownCommandError):
            _shutdown()

    def test_replace_stored(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.replace(b'key', b'value', noreply=False)
        assert result is True

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.replace(b'key', b'value', noreply=False)
        assert result is True

    def test_replace_not_stored(self):
        client = self.make_client([b'NOT_STORED\r\n'])
        result = client.replace(b'key', b'value', noreply=False)
        assert result is False

        # unit test for encoding passed in __init__
        client = self.make_client([b'NOT_STORED\r\n'], encoding='utf-8')
        result = client.replace(b'key', b'value', noreply=False)
        assert result is False

    def test_serialization(self):
        class JsonSerde(object):
            def serialize(self, key, value):
                return json.dumps(value).encode('ascii'), 0

            def deserialize(self, key, value, flags):
                return json.loads(value.decode('ascii'))

        client = self.make_client([b'STORED\r\n'], serde=JsonSerde())
        client.set('key', {'c': 'd'})
        assert client.sock.send_bufs == [
            b'set key 0 0 10 noreply\r\n{"c": "d"}\r\n'
        ]

    def test_serialization_flags(self):
        def _ser(key, value):
            return value, 1 if isinstance(value, int) else 0

        client = self.make_client(
            [b'STORED\r\n', b'STORED\r\n'], serializer=_ser)
        client.set_many(
            collections.OrderedDict([(b'a', b's'), (b'b', 0)]), noreply=False)
        assert client.sock.send_bufs == [
            b'set a 0 0 1\r\ns\r\nset b 1 0 1\r\n0\r\n'
        ]

    def test_serialization_overridden_flags(self):
        def _ser(key, value):
            return value, 1 if isinstance(value, int) else 0

        client = self.make_client(
            [b'STORED\r\n', b'STORED\r\n'], serializer=_ser)
        client.set_many(
            collections.OrderedDict([(b'a', b's'), (b'b', 0)]),
            noreply=False, flags=5)
        assert client.sock.send_bufs == [
            b'set a 5 0 1\r\ns\r\nset b 5 0 1\r\n0\r\n'
        ]

    def test_explicit_flags(self):
        client = self.make_client([b'STORED\r\n', b'STORED\r\n'])
        client.set_many(
            collections.OrderedDict([(b'a', b's'), (b'b', 0)]),
            noreply=False, flags=5)
        assert client.sock.send_bufs == [
            b'set a 5 0 1\r\ns\r\nset b 5 0 1\r\n0\r\n'
        ]

    def test_set_socket_handling(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.set(b'key', b'value', noreply=False)
        assert result is True
        assert client.sock.closed is False
        assert len(client.sock.send_bufs) == 1

    def test_set_error(self):
        client = self.make_client([b'ERROR\r\n'])

        def _set():
            client.set(b'key', b'value', noreply=False)

        with pytest.raises(MemcacheUnknownCommandError):
            _set()

    def test_set_exception(self):
        client = self.make_client([Exception('fail')])

        def _set():
            client.set(b'key', b'value', noreply=False)
        with pytest.raises(Exception):
            _set()

        assert client.sock is None

    def test_set_client_error(self):
        client = self.make_client([b'CLIENT_ERROR some message\r\n'])

        def _set():
            client.set('key', 'value', noreply=False)

        with pytest.raises(MemcacheClientError):
            _set()

    def test_set_server_error(self):
        client = self.make_client([b'SERVER_ERROR some message\r\n'])

        def _set():
            client.set(b'key', b'value', noreply=False)

        with pytest.raises(MemcacheServerError):
            _set()

    def test_closing_socket_on_unexpected_closed_error(self):
        client = self.make_client([
            b'VALUE ',
            MemcacheUnexpectedCloseError("foo bar"),
        ])

        def _set():
            client.set(b'key', b'value', noreply=False)

        with pytest.raises(MemcacheUnexpectedCloseError):
            _set()

        assert client.sock is None

    def test_set_unknown_error(self):
        client = self.make_client([b'foobarbaz\r\n'])

        def _set():
            client.set(b'key', b'value', noreply=False)

        with pytest.raises(MemcacheUnknownError):
            _set()

    def test_set_key_with_space(self):
        client = self.make_client([b''])

        def _set():
            client.set(b'key has space', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_key_with_newline(self):
        client = self.make_client([b''])

        def _set():
            client.set(b'key\n', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_key_with_carriage_return(self):
        client = self.make_client([b''])

        def _set():
            client.set(b'key\r', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_key_with_null_character(self):
        client = self.make_client([b''])

        def _set():
            client.set(b'key\00', b'value', noreply=False)

        with pytest.raises(MemcacheIllegalInputError):
            _set()

    def test_set_key_with_noninteger_expire(self):
        client = self.make_client([b''])

        class _OneLike(object):
            """object that looks similar to the int 1"""
            def __str__(self):
                return "1"

        for noreply in (True, False):
            for expire in (1.5, _OneLike(), "1"):
                def _set():
                    client.set(b'finekey', b'finevalue',
                               noreply=noreply, expire=expire)

                with pytest.raises(MemcacheIllegalInputError):
                    _set()

    def test_set_many_socket_handling(self):
        client = self.make_client([b'STORED\r\n'])
        result = client.set_many({b'key': b'value'}, noreply=False)
        assert result == []
        assert client.sock.closed is False
        assert len(client.sock.send_bufs) == 1

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n'], encoding='utf-8')
        result = client.set_many({b'key': b'value'}, noreply=False)
        assert result == []
        assert client.sock.closed is False
        assert len(client.sock.send_bufs) == 1

    def test_set_many_exception(self):
        client = self.make_client([b'STORED\r\n', Exception('fail')])

        def _set():
            client.set_many({b'key': b'value', b'other': b'value'},
                            noreply=False)

        with pytest.raises(Exception):
            _set()

        assert client.sock is None

        # unit test for encoding passed in __init__()
        client = self.make_client([b'STORED\r\n', Exception('fail')],
                                  encoding='utf-8')

        def _set():
            client.set_many({b'key': b'value', b'other': b'value'},
                            noreply=False)

        with pytest.raises(Exception):
            _set()

        assert client.sock is None

    def test_stats(self):
        client = self.make_client([b'STAT fake_stats 1\r\n', b'END\r\n'])
        result = client.stats()
        assert client.sock.send_bufs == [
            b'stats\r\n'
        ]
        assert result == {b'fake_stats': 1}

    def test_stats_with_args(self):
        client = self.make_client([b'STAT fake_stats 1\r\n', b'END\r\n'])
        result = client.stats('some_arg')
        assert client.sock.send_bufs == [
            b'stats some_arg\r\n'
        ]
        assert result == {b'fake_stats': 1}

    def test_stats_conversions(self):
        client = self.make_client([
            # Most stats are converted to int
            b'STAT cmd_get 2519\r\n',
            b'STAT cmd_set 3099\r\n',
            b'STAT evictions 939\r\n',

            # Unless they can't be, they remain str
            b'STAT libevent 2.0.19-stable\r\n',

            # Some named stats are explicitly converted
            b'STAT hash_is_expanding 0\r\n',
            b'STAT rusage_user 0.609165\r\n',
            b'STAT rusage_system 0.852791\r\n',
            b'STAT slab_reassign_running 1\r\n',
            b'STAT version 1.4.14\r\n',
            b'STAT umask 777\r\n',
            b'STAT auth_enabled_sasl yes\r\n',
            b'END\r\n',
        ])
        result = client.stats()
        assert client.sock.send_bufs == [
            b'stats\r\n'
        ]
        expected = {
            b'cmd_get': 2519,
            b'cmd_set': 3099,
            b'evictions': 939,
            b'libevent': b'2.0.19-stable',
            b'hash_is_expanding': False,
            b'rusage_user': 0.609165,
            b'rusage_system': 0.852791,
            b'slab_reassign_running': True,
            b'version': b'1.4.14',
            b'umask': 0o777,
            b'auth_enabled_sasl': True,
        }
        assert result == expected

    def test_stats_cachedump(self):
        client = self.make_client([b'ITEM bob [7 b; 0 s]\r\n', b'END\r\n'])
        result = client.stats('cachedump', '1', '1')
        assert client.sock.send_bufs == [
            b'stats cachedump 1 1\r\n'
        ]
        assert result == {b'bob': b'[7 b; 0 s]'}

    def test_cache_memlimit(self):
        client = self.make_client([b'OK\r\n'])
        result = client.cache_memlimit(8)
        assert client.sock.send_bufs == [
            b'cache_memlimit 8\r\n'
        ]
        assert result is True

    def test_python_dict_set_is_supported(self):
        client = self.make_client([b'STORED\r\n'])
        client[b'key'] = b'value'

    def test_python_dict_get_is_supported(self):
        client = self.make_client([b'VALUE key 0 5\r\nvalue\r\nEND\r\n'])
        assert client[b'key'] == b'value'

    def test_python_dict_get_not_found_is_supported(self):
        client = self.make_client([b'END\r\n'])

        def _get():
            client[b'key']

        with pytest.raises(KeyError):
            _get()

    def test_python_dict_del_is_supported(self):
        client = self.make_client([b'DELETED\r\n'])
        del client[b'key']

    def test_too_long_key(self):
        client = self.make_client([b'END\r\n'])

        with pytest.raises(MemcacheClientError):
            client.get(b'x' * 251)

    def test_too_long_unicode_key(self):
        client = self.make_client([b'STORED\r\n'], allow_unicode_keys=True)

        with pytest.raises(MemcacheClientError):
            client.get('my☃'*150)

        with pytest.raises(MemcacheClientError):
            client.get(u'\u0FFF'*150)

    def test_key_contains_space(self):
        client = self.make_client([b'END\r\n'])
        with pytest.raises(MemcacheClientError):
            client.get(b'abc xyz')

    def test_key_contains_nonascii(self):
        client = self.make_client([b'END\r\n'])

        with pytest.raises(MemcacheClientError):
            client.get(u'\u3053\u3093\u306b\u3061\u306f')

    def _default_noreply_false(self, cmd, args, response):
        client = self.make_client(response, default_noreply=False)
        result = getattr(client, cmd)(*args)
        assert result is False

    def _default_noreply_true(self, cmd, args, response):
        client = self.make_client(response, default_noreply=True)
        result = getattr(client, cmd)(*args)
        assert result is True

    def _default_noreply_true_and_empty_list(self, cmd, args, response):
        client = self.make_client(response, default_noreply=True)
        result = getattr(client, cmd)(*args)
        assert result == []

    def test_default_noreply_set(self):
        with pytest.raises(MemcacheUnknownError):
            self._default_noreply_false(
                'set', (b'key', b'value'), [b'UNKNOWN\r\n'])
        self._default_noreply_false(
            'set', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'set', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_set_many(self):
        with pytest.raises(MemcacheUnknownError):
            client = self.make_client([b'UNKNOWN\r\n'], default_noreply=False)
            result = client.set_many({b'key': b'value'})
            assert result == [b'key']
        self._default_noreply_true_and_empty_list(
            'set_many', ({b'key': b'value'},), [b'NOT_STORED\r\n'])

    def test_default_noreply_add(self):
        self._default_noreply_false(
            'add', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'add', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_replace(self):
        self._default_noreply_false(
            'replace', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'replace', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_append(self):
        self._default_noreply_false(
            'append', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'append', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_prepend(self):
        self._default_noreply_false(
            'prepend', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'prepend', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_touch(self):
        self._default_noreply_false('touch', (b'key',), [b'NOT_FOUND\r\n'])
        self._default_noreply_true('touch', (b'key',), [b'NOT_FOUND\r\n'])

    def test_default_noreply_flush_all(self):
        self._default_noreply_false('flush_all', (),
                                    [b'__FAKE_RESPONSE__\r\n'])
        self._default_noreply_true('flush_all', (), [b'__FAKE_RESPONSE__\r\n'])

    def test_version_success(self):
        client = self.make_client([b'VERSION 1.2.3\r\n'],
                                  default_noreply=False)
        result = client.version()
        assert result == b'1.2.3'

    def test_version_exception(self):
        client = self.make_client([b'INVALID DATA\r\n'], default_noreply=False)
        with pytest.raises(MemcacheUnknownError):
            client.version()


@pytest.mark.unit()
class TestClientSocketConnect(unittest.TestCase):
    def test_socket_connect_ipv4(self):
        server = ('127.0.0.1', 11211)

        client = Client(server, socket_module=MockSocketModule())
        client._connect()
        assert client.sock.connections == [server]
        assert client.sock.family == socket.AF_INET

        timeout = 2
        connect_timeout = 3
        client = Client(
            server, connect_timeout=connect_timeout, timeout=timeout,
            socket_module=MockSocketModule())
        client._connect()
        assert client.sock.timeouts == [connect_timeout, timeout]

        client = Client(server, socket_module=MockSocketModule())
        client._connect()
        assert client.sock.socket_options == []

        client = Client(
            server, socket_module=MockSocketModule(), no_delay=True)
        client._connect()
        assert client.sock.socket_options == [(socket.IPPROTO_TCP,
                                               socket.TCP_NODELAY, 1)]

    def test_socket_connect_ipv6(self):
        server = ('::1', 11211)

        client = Client(server, socket_module=MockSocketModule())
        client._connect()
        assert client.sock.connections == [server + (0, 0)]
        assert client.sock.family == socket.AF_INET6

        timeout = 2
        connect_timeout = 3
        client = Client(
            server, connect_timeout=connect_timeout, timeout=timeout,
            socket_module=MockSocketModule())
        client._connect()
        assert client.sock.timeouts == [connect_timeout, timeout]

        client = Client(server, socket_module=MockSocketModule())
        client._connect()
        assert client.sock.socket_options == []

        client = Client(
            server, socket_module=MockSocketModule(), no_delay=True)
        client._connect()
        assert client.sock.socket_options == [(socket.IPPROTO_TCP,
                                               socket.TCP_NODELAY, 1)]

    def test_socket_connect_unix(self):
        server = '/tmp/pymemcache.{pid}'.format(pid=os.getpid())

        with MockUnixSocketServer(server):
            client = Client(server)
            client._connect()
            assert client.sock.family == socket.AF_UNIX

    @unittest.skipIf('Linux' != platform.system(),
                     'Socket keepalive only support Linux platforms.')
    def test_linux_socket_keepalive(self):
        server = ('::1', 11211)
        try:
            client = Client(
                server,
                socket_module=MockSocketModule(),
                socket_keepalive=KeepaliveOpts())
            client._connect()
        except SystemError:
            self.fail("SystemError unexpectedly raised")
        with self.assertRaises(ValueError):
            # A KeepaliveOpts object is expected, a ValueError will be raised
            Client(
                server,
                socket_module=MockSocketModule(),
                socket_keepalive=True)

    @mock.patch('platform.system')
    def test_osx_socket_keepalive(self, platform_mock):
        platform_mock.return_value = 'Darwin'
        server = ('::1', 11211)
        # For the moment the socket keepalive is only implemented for Linux
        with self.assertRaises(SystemError):
            Client(
                server,
                socket_module=MockSocketModule(),
                socket_keepalive=KeepaliveOpts())

    @mock.patch('platform.system')
    def test_windows_socket_keepalive(self, platform_mock):
        platform_mock.return_value = 'Windows'
        server = ('::1', 11211)
        # For the moment the socket keepalive is only implemented for Linux
        with self.assertRaises(SystemError):
            Client(
                server,
                socket_module=MockSocketModule(),
                socket_keepalive=KeepaliveOpts())

    def test_socket_connect_closes_on_failure(self):
        server = ("example.com", 11211)

        socket_module = MockSocketModule(connect_failure=OSError())
        client = Client(server, socket_module=socket_module)
        with pytest.raises(OSError):
            client._connect()
        assert len(socket_module.sockets) == 1
        assert socket_module.sockets[0].connections == []
        assert socket_module.sockets[0].closed

    def test_socket_close(self):
        server = ("example.com", 11211)

        client = Client(server, socket_module=MockSocketModule())
        client._connect()
        assert client.sock is not None

        client.close()
        assert client.sock is None

    def test_socket_close_exception(self):
        server = ("example.com", 11211)

        socket_module = MockSocketModule(close_failure=OSError())
        client = Client(server, socket_module=socket_module)
        client._connect()
        assert client.sock is not None

        client.close()
        assert client.sock is None


class TestPooledClient(ClientTestMixin, unittest.TestCase):
    def make_client(self, mock_socket_values, **kwargs):
        mock_client = Client(None, **kwargs)
        mock_client.sock = MockSocket(list(mock_socket_values))
        client = PooledClient(None, **kwargs)
        client.client_pool = pool.ObjectPool(lambda: mock_client)
        return client

    def _default_noreply_false(self, cmd, args, response):
        client = self.make_client(response, default_noreply=False)
        result = getattr(client, cmd)(*args)
        assert result is False

    def _default_noreply_true(self, cmd, args, response):
        client = self.make_client(response, default_noreply=True)
        result = getattr(client, cmd)(*args)
        assert result is True

    def _default_noreply_true_and_empty_list(self, cmd, args, response):
        client = self.make_client(response, default_noreply=True)
        result = getattr(client, cmd)(*args)
        assert result == []

    def test_default_noreply_set(self):
        with pytest.raises(MemcacheUnknownError):
            self._default_noreply_false(
                'set', (b'key', b'value'), [b'UNKNOWN\r\n'])
        self._default_noreply_false(
            'set', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'set', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_set_many(self):
        with pytest.raises(MemcacheUnknownError):
            client = self.make_client([b'UNKNOWN\r\n'], default_noreply=False)
            client.set_many({b'key': b'value'})
        self._default_noreply_true_and_empty_list(
            'set_many', ({b'key': b'value'},), [b'NOT_STORED\r\n'])

    def test_default_noreply_add(self):
        self._default_noreply_false(
            'add', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'add', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_replace(self):
        self._default_noreply_false(
            'replace', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'replace', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_append(self):
        self._default_noreply_false(
            'append', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'append', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_prepend(self):
        self._default_noreply_false(
            'prepend', (b'key', b'value'), [b'NOT_STORED\r\n'])
        self._default_noreply_true(
            'prepend', (b'key', b'value'), [b'NOT_STORED\r\n'])

    def test_default_noreply_touch(self):
        self._default_noreply_false('touch', (b'key',), [b'NOT_FOUND\r\n'])
        self._default_noreply_true('touch', (b'key',), [b'NOT_FOUND\r\n'])

    def test_default_noreply_flush_all(self):
        self._default_noreply_false('flush_all', (),
                                    [b'__FAKE_RESPONSE__\r\n'])
        self._default_noreply_true('flush_all', (), [b'__FAKE_RESPONSE__\r\n'])

    def test_custom_client(self):
        class MyClient(Client):
            pass
        client = PooledClient(('host', 11211))
        client.client_class = MyClient
        assert isinstance(client.client_pool.get(), MyClient)


class TestPooledClientIdleTimeout(ClientTestMixin, unittest.TestCase):
    def make_client(self, mock_socket_values, **kwargs):
        mock_client = Client(None, **kwargs)
        mock_client.sock = MockSocket(list(mock_socket_values))
        client = PooledClient(None, pool_idle_timeout=60, **kwargs)
        client.client_pool = pool.ObjectPool(lambda: mock_client)
        return client

    def test_free_idle(self):
        class Counter(object):
            count = 0

            def increment(self, obj):
                self.count += 1

        removed = Counter()

        client = self.make_client([b'VALUE key 0 5\r\nvalue\r\nEND\r\n']*2)
        client.client_pool._after_remove = removed.increment
        client.client_pool._idle_clock = lambda: 0

        client.set(b'key', b'value')
        assert removed.count == 0
        client.get(b'key')
        assert removed.count == 0

        # Advance clock to beyond the idle timeout.
        client.client_pool._idle_clock = lambda: 61
        client.get(b'key')
        assert removed.count == 1


class TestMockClient(ClientTestMixin, unittest.TestCase):
    def make_client(self, mock_socket_values, **kwargs):
        client = MockMemcacheClient(None, **kwargs)
        client.sock = MockSocket(list(mock_socket_values))
        return client

    def test_get_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'VALUE key 0 5\r\nvalue\r\nEND\r\n',
        ])
        result = client.set(b'key', b'value', noreply=False)
        result = client.get(b'key')
        assert result == b'value'

    def test_deserialization(self):
        class JsonSerde(object):
            def serialize(self, key, value):
                if isinstance(value, dict):
                    return json.dumps(value).encode('UTF-8'), 1
                return value, 0

            def deserialize(self, key, value, flags):
                if flags == 1:
                    return json.loads(value.decode('UTF-8'))
                return value

        client = self.make_client([
            b'STORED\r\n',
            b'VALUE key1 0 5\r\nhello\r\nEND\r\n',
            b'STORED\r\n',
            b'VALUE key2 0 18\r\n{"hello": "world"}\r\nEND\r\n',
        ], serde=JsonSerde())

        result = client.set(b'key1', b'hello', noreply=False)
        result = client.get(b'key1')
        assert result == b'hello'

        result = client.set(b'key2', dict(hello='world'), noreply=False)
        result = client.get(b'key2')
        assert result == dict(hello='world')


class TestPrefixedClient(ClientTestMixin, unittest.TestCase):
    def make_client(self, mock_socket_values, **kwargs):
        client = Client(None, key_prefix=b'xyz:', **kwargs)
        client.sock = MockSocket(list(mock_socket_values))
        return client

    def test_get_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'VALUE xyz:key 0 5\r\nvalue\r\nEND\r\n',
        ])
        result = client.set(b'key', b'value', noreply=False)
        result = client.get(b'key')
        assert result == b'value'

    def test_get_many_some_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'VALUE xyz:key1 0 6\r\nvalue1\r\nEND\r\n',
        ])
        result = client.set(b'key1', b'value1', noreply=False)
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1'}

    def test_get_many_all_found(self):
        client = self.make_client([
            b'STORED\r\n',
            b'STORED\r\n',
            b'VALUE xyz:key1 0 6\r\nvalue1\r\n',
            b'VALUE xyz:key2 0 6\r\nvalue2\r\nEND\r\n',
        ])
        result = client.set(b'key1', b'value1', noreply=False)
        result = client.set(b'key2', b'value2', noreply=False)
        result = client.get_many([b'key1', b'key2'])
        assert result == {b'key1': b'value1', b'key2': b'value2'}

    def test_python_dict_get_is_supported(self):
        client = self.make_client([b'VALUE xyz:key 0 5\r\nvalue\r\nEND\r\n'])
        assert client[b'key'] == b'value'


class TestPrefixedPooledClient(TestPrefixedClient):
    def make_client(self, mock_socket_values, **kwargs):
        mock_client = Client(None, key_prefix=b'xyz:', **kwargs)
        mock_client.sock = MockSocket(list(mock_socket_values))
        client = PooledClient(None, key_prefix=b'xyz:', **kwargs)
        client.client_pool = pool.ObjectPool(lambda: mock_client)
        return client


@pytest.mark.unit()
class TestRetryOnEINTR(unittest.TestCase):
    def make_client(self, values):
        client = Client(None)
        client.sock = MockSocket(list(values))
        return client

    def test_recv(self):
        client = self.make_client([
            b'VALUE ',
            socket.error(errno.EINTR, "Interrupted system call"),
            b'key1 0 6\r\nval',
            socket.error(errno.EINTR, "Interrupted system call"),
            b'ue1\r\nEND\r\n',
        ])
        assert client[b'key1'] == b'value1'


@pytest.mark.unit()
class TestNormalizeServerSpec(unittest.TestCase):
    def test_normalize_server_spec(self):
        f = normalize_server_spec
        assert f(None) is None
        assert f(('127.0.0.1', 12345)) == ('127.0.0.1', 12345)
        assert f(['127.0.0.1', 12345]) == ('127.0.0.1', 12345)
        assert f('unix:/run/memcached/socket') == '/run/memcached/socket'
        assert f('/run/memcached/socket') == '/run/memcached/socket'
        assert f('localhost') == ('localhost', 11211)
        assert f('localhost:12345') == ('localhost', 12345)
        assert f('[::1]') == ('::1', 11211)
        assert f('[::1]:12345') == ('::1', 12345)
        assert f('127.0.0.1') == ('127.0.0.1', 11211)
        assert f('127.0.0.1:12345') == ('127.0.0.1', 12345)

        with pytest.raises(ValueError) as excinfo:
            f({'host': 12345})
        assert str(excinfo.value) == "Unknown server provided: {'host': 12345}"

        with pytest.raises(ValueError) as excinfo:
            f(12345)
        assert str(excinfo.value) == "Unknown server provided: 12345"


@pytest.mark.unit()
class TestKeepaliveopts(unittest.TestCase):
    def test_keepalive_opts(self):
        kao = KeepaliveOpts()
        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 5)
        kao = KeepaliveOpts(idle=1, intvl=4, cnt=2)
        assert (kao.idle == 1 and kao.intvl == 4 and kao.cnt == 2)
        kao = KeepaliveOpts(idle=8)
        assert (kao.idle == 8 and kao.intvl == 1 and kao.cnt == 5)
        kao = KeepaliveOpts(cnt=8)
        assert (kao.idle == 1 and kao.intvl == 1 and kao.cnt == 8)

        with self.assertRaises(ValueError):
            KeepaliveOpts(cnt=0)
        with self.assertRaises(ValueError):
            KeepaliveOpts(idle=-1)



""" Test collection for the RetryingClient. """

import functools
import unittest

import mock
import pytest

from .test_client import ClientTestMixin, MockSocket
from pymemcache.client.retrying import RetryingClient
from pymemcache.client.base import Client
from pymemcache.exceptions import MemcacheUnknownError, MemcacheClientError


# Test pure passthroughs with no retry action.
class TestRetryingClientPassthrough(ClientTestMixin, unittest.TestCase):

    def make_base_client(self, mock_socket_values, **kwargs):
        base_client = Client(None, **kwargs)
        # mock out client._connect() rather than hard-settting client.sock to
        # ensure methods are checking whether self.sock is None before
        # attempting to use it
        sock = MockSocket(list(mock_socket_values))
        base_client._connect = mock.Mock(side_effect=functools.partial(
            setattr, base_client, "sock", sock))
        return base_client

    def make_client(self, mock_socket_values, **kwargs):
        # Create a base client to wrap.
        base_client = self.make_base_client(
            mock_socket_values=mock_socket_values, **kwargs
        )

        # Wrap the client in the retrying class, disable retries.
        client = RetryingClient(base_client, attempts=1)
        return client


# Retry specific tests.
@pytest.mark.unit()
class TestRetryingClient(object):

    def make_base_client(self, mock_socket_values, **kwargs):
        """ Creates a regular mock client to wrap in the RetryClient. """
        base_client = Client(None, **kwargs)
        # mock out client._connect() rather than hard-settting client.sock to
        # ensure methods are checking whether self.sock is None before
        # attempting to use it
        sock = MockSocket(list(mock_socket_values))
        base_client._connect = mock.Mock(side_effect=functools.partial(
            setattr, base_client, "sock", sock))
        return base_client

    def make_client(self, mock_socket_values, **kwargs):
        """
        Creates a RetryingClient that will respond with the given values,
        configured using kwargs.
        """
        # Create a base client to wrap.
        base_client = self.make_base_client(
            mock_socket_values=mock_socket_values
        )

        # Wrap the client in the retrying class, and pass kwargs on.
        client = RetryingClient(base_client, **kwargs)
        return client

    # Start testing.
    def test_constructor_default(self):
        base_client = self.make_base_client([])
        RetryingClient(base_client)

        with pytest.raises(TypeError):
            RetryingClient()

    def test_constructor_attempts(self):
        base_client = self.make_base_client([])
        rc = RetryingClient(base_client, attempts=1)
        assert rc._attempts == 1

        with pytest.raises(ValueError):
            RetryingClient(base_client, attempts=0)

    def test_constructor_retry_for(self):
        base_client = self.make_base_client([])

        # Try none/default.
        rc = RetryingClient(base_client, retry_for=None)
        assert rc._retry_for == tuple()

        # Try with tuple.
        rc = RetryingClient(base_client, retry_for=tuple([Exception]))
        assert rc._retry_for == tuple([Exception])

        # Try with list.
        rc = RetryingClient(base_client, retry_for=[Exception])
        assert rc._retry_for == tuple([Exception])

        # Try with multi element list.
        rc = RetryingClient(base_client, retry_for=[Exception, IOError])
        assert rc._retry_for == (Exception, IOError)

        # With string?
        with pytest.raises(ValueError):
            RetryingClient(base_client, retry_for="haha!")

        # With collectino of string and exceptions?
        with pytest.raises(ValueError):
            RetryingClient(base_client, retry_for=[Exception, str])

    def test_constructor_do_no_retry_for(self):
        base_client = self.make_base_client([])

        # Try none/default.
        rc = RetryingClient(base_client, do_not_retry_for=None)
        assert rc._do_not_retry_for == tuple()

        # Try with tuple.
        rc = RetryingClient(base_client, do_not_retry_for=tuple([Exception]))
        assert rc._do_not_retry_for == tuple([Exception])

        # Try with list.
        rc = RetryingClient(base_client, do_not_retry_for=[Exception])
        assert rc._do_not_retry_for == tuple([Exception])

        # Try with multi element list.
        rc = RetryingClient(base_client, do_not_retry_for=[Exception, IOError])
        assert rc._do_not_retry_for == (Exception, IOError)

        # With string?
        with pytest.raises(ValueError):
            RetryingClient(base_client, do_not_retry_for="haha!")

        # With collectino of string and exceptions?
        with pytest.raises(ValueError):
            RetryingClient(base_client, do_not_retry_for=[Exception, str])

    def test_constructor_both_filters(self):
        base_client = self.make_base_client([])

        # Try none/default.
        rc = RetryingClient(base_client, retry_for=None, do_not_retry_for=None)
        assert rc._retry_for == tuple()
        assert rc._do_not_retry_for == tuple()

        # Try a valid config.
        rc = RetryingClient(
            base_client,
            retry_for=[Exception, IOError],
            do_not_retry_for=[ValueError, MemcacheUnknownError]
        )
        assert rc._retry_for == (Exception, IOError)
        assert rc._do_not_retry_for == (ValueError, MemcacheUnknownError)

        # Try with overlapping filters
        with pytest.raises(ValueError):
            rc = RetryingClient(
                base_client,
                retry_for=[Exception, IOError, MemcacheUnknownError],
                do_not_retry_for=[ValueError, MemcacheUnknownError]
            )

    def test_dir_passthrough(self):
        base = self.make_base_client([])
        client = RetryingClient(base)

        assert dir(base) == dir(client)

    def test_retry_dict_set_is_supported(self):
        client = self.make_client([b'UNKNOWN\r\n', b'STORED\r\n'])
        client[b'key'] = b'value'

    def test_retry_dict_get_is_supported(self):
        client = self.make_client(
            [
                b'UNKNOWN\r\n',
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
            ]
        )
        assert client[b'key'] == b'value'

    def test_retry_dict_get_not_found_is_supported(self):
        client = self.make_client([b'UNKNOWN\r\n', b'END\r\n'])

        with pytest.raises(KeyError):
            client[b'key']

    def test_retry_dict_del_is_supported(self):
        client = self.make_client([b'UNKNOWN\r\n', b'DELETED\r\n'])
        del client[b'key']

    def test_retry_get_found(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
        ], attempts=2)
        result = client.get("key")
        assert result == b'value'

    def test_retry_get_not_found(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'END\r\n'
        ], attempts=2)
        result = client.get("key")
        assert result is None

    def test_retry_get_exception(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'UNKNOWN\r\n'
        ], attempts=2)
        with pytest.raises(MemcacheUnknownError):
            client.get("key")

    def test_retry_set_success(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'STORED\r\n'
        ], attempts=2)
        result = client.set("key", "value", noreply=False)
        assert result is True

    def test_retry_set_fail(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'UNKNOWN\r\n',
            b'STORED\r\n'
        ], attempts=2)
        with pytest.raises(MemcacheUnknownError):
            client.set("key", "value", noreply=False)

    def test_no_retry(self):
        client = self.make_client([
            b'UNKNOWN\r\n',
            b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
        ], attempts=1)

        with pytest.raises(MemcacheUnknownError):
            client.get("key")

    def test_retry_for_exception_success(self):
        # Test that we retry for the exception specified.
        client = self.make_client(
            [
                MemcacheClientError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
            ],
            attempts=2,
            retry_for=tuple([MemcacheClientError])
        )
        result = client.get("key")
        assert result == b'value'

    def test_retry_for_exception_fail(self):
        # Test that we do not retry for unapproved exception.
        client = self.make_client(
            [
                MemcacheUnknownError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
            ],
            attempts=2,
            retry_for=tuple([MemcacheClientError])
        )

        with pytest.raises(MemcacheUnknownError):
            client.get("key")

    def test_do_not_retry_for_exception_success(self):
        # Test that we retry for exceptions not specified.
        client = self.make_client(
            [
                MemcacheClientError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
            ],
            attempts=2,
            do_not_retry_for=tuple([MemcacheUnknownError])
        )
        result = client.get("key")
        assert result == b'value'

    def test_do_not_retry_for_exception_fail(self):
        # Test that we do not retry for the exception specified.
        client = self.make_client(
            [
                MemcacheClientError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n'
            ],
            attempts=2,
            do_not_retry_for=tuple([MemcacheClientError])
        )

        with pytest.raises(MemcacheClientError):
            client.get("key")

    def test_both_exception_filters(self):
        # Test interacction between both exception filters.
        client = self.make_client(
            [
                MemcacheClientError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n',
                MemcacheUnknownError("Whoops."),
                b'VALUE key 0 5\r\nvalue\r\nEND\r\n',
            ],
            attempts=2,
            retry_for=tuple([MemcacheClientError]),
            do_not_retry_for=tuple([MemcacheUnknownError])
        )

        # Check that we succeed where allowed.
        result = client.get("key")
        assert result == b'value'

        # Check that no retries are attempted for the banned exception.
        with pytest.raises(MemcacheUnknownError):
            client.get("key")



class MemcacheError(Exception):
    "Base exception class"
    pass


class MemcacheClientError(MemcacheError):
    """Raised when memcached fails to parse the arguments to a request, likely
    due to a malformed key and/or value, a bug in this library, or a version
    mismatch with memcached."""
    pass


class MemcacheUnknownCommandError(MemcacheClientError):
    """Raised when memcached fails to parse a request, likely due to a bug in
    this library or a version mismatch with memcached."""
    pass


class MemcacheIllegalInputError(MemcacheClientError):
    """Raised when a key or value is not legal for Memcache (see the class docs
    for Client for more details)."""
    pass


class MemcacheServerError(MemcacheError):
    """Raised when memcached reports a failure while processing a request,
    likely due to a bug or transient issue in memcached."""
    pass


class MemcacheUnknownError(MemcacheError):
    """Raised when this library receives a response from memcached that it
    cannot parse, likely due to a bug in this library or a version mismatch
    with memcached."""
    pass


class MemcacheUnexpectedCloseError(MemcacheServerError):
    "Raised when the connection with memcached closes unexpectedly."
    pass



# Copyright 2015 Yahoo.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections
import contextlib
import sys
import threading
import time

import six


class ObjectPool(object):
    """A pool of objects that release/creates/destroys as needed."""

    def __init__(self, obj_creator,
                 after_remove=None, max_size=None,
                 idle_timeout=0,
                 lock_generator=None):
        self._used_objs = collections.deque()
        self._free_objs = collections.deque()
        self._obj_creator = obj_creator
        if lock_generator is None:
            self._lock = threading.Lock()
        else:
            self._lock = lock_generator()
        self._after_remove = after_remove
        max_size = max_size or 2 ** 31
        if not isinstance(max_size, six.integer_types) or max_size < 0:
            raise ValueError('"max_size" must be a positive integer')
        self.max_size = max_size
        self.idle_timeout = idle_timeout
        self._idle_clock = time.time if idle_timeout else int

    @property
    def used(self):
        return tuple(self._used_objs)

    @property
    def free(self):
        return tuple(self._free_objs)

    @contextlib.contextmanager
    def get_and_release(self, destroy_on_fail=False):
        obj = self.get()
        try:
            yield obj
        except Exception:
            exc_info = sys.exc_info()
            if not destroy_on_fail:
                self.release(obj)
            else:
                self.destroy(obj)
            six.reraise(exc_info[0], exc_info[1], exc_info[2])
        self.release(obj)

    def get(self):
        with self._lock:
            # Find a free object, removing any that have idled for too long.
            now = self._idle_clock()
            while self._free_objs:
                obj = self._free_objs.popleft()
                if now - obj._last_used <= self.idle_timeout:
                    break

                if self._after_remove is not None:
                    self._after_remove(obj)
            else:
                # No free objects, create a new one.
                curr_count = len(self._used_objs)
                if curr_count >= self.max_size:
                    raise RuntimeError("Too many objects,"
                                       " %s >= %s" % (curr_count,
                                                      self.max_size))
                obj = self._obj_creator()

            self._used_objs.append(obj)
            obj._last_used = now
            return obj

    def destroy(self, obj, silent=True):
        was_dropped = False
        with self._lock:
            try:
                self._used_objs.remove(obj)
                was_dropped = True
            except ValueError:
                if not silent:
                    raise
        if was_dropped and self._after_remove is not None:
            self._after_remove(obj)

    def release(self, obj, silent=True):
        with self._lock:
            try:
                self._used_objs.remove(obj)
                self._free_objs.append(obj)
                obj._last_used = self._idle_clock()
            except ValueError:
                if not silent:
                    raise

    def clear(self):
        if self._after_remove is not None:
            needs_destroy = []
            with self._lock:
                needs_destroy.extend(self._used_objs)
                needs_destroy.extend(self._free_objs)
                self._free_objs.clear()
                self._used_objs.clear()
            for obj in needs_destroy:
                self._after_remove(obj)
        else:
            with self._lock:
                self._free_objs.clear()
                self._used_objs.clear()



# -*- coding: utf-8 -*-
from unittest import TestCase

from pymemcache.serde import (pickle_serde,
                              PickleSerde,
                              FLAG_BYTES,
                              FLAG_PICKLE, FLAG_INTEGER, FLAG_LONG, FLAG_TEXT)
import pytest
import six
from six.moves import cPickle as pickle


class CustomInt(int):
    """
    Custom integer type for testing.

    Entirely useless, but used to show that built in types get serialized and
    deserialized back as the same type of object.
    """
    pass


@pytest.mark.unit()
class TestSerde(TestCase):
    serde = pickle_serde

    def check(self, value, expected_flags):
        serialized, flags = self.serde.serialize(b'key', value)
        assert flags == expected_flags

        # pymemcache stores values as byte strings, so we immediately the value
        # if needed so deserialized works as it would with a real server
        if not isinstance(serialized, six.binary_type):
            serialized = six.text_type(serialized).encode('ascii')

        deserialized = self.serde.deserialize(b'key', serialized, flags)
        assert deserialized == value

    def test_bytes(self):
        self.check(b'value', FLAG_BYTES)
        self.check(b'\xc2\xa3 $ \xe2\x82\xac', FLAG_BYTES)  # £ $ €

    def test_unicode(self):
        self.check(u'value', FLAG_TEXT)
        self.check(u'£ $ €', FLAG_TEXT)

    def test_int(self):
        self.check(1, FLAG_INTEGER)

    def test_long(self):
        # long only exists with Python 2, so we're just testing for another
        # integer with Python 3
        if six.PY2:
            expected_flags = FLAG_LONG
        else:
            expected_flags = FLAG_INTEGER
        self.check(123123123123123123123, expected_flags)

    def test_pickleable(self):
        self.check({'a': 'dict'}, FLAG_PICKLE)

    def test_subtype(self):
        # Subclass of a native type will be restored as the same type
        self.check(CustomInt(123123), FLAG_PICKLE)


@pytest.mark.unit()
class TestSerdePickleVersion0(TestCase):
    serde = PickleSerde(pickle_version=0)


@pytest.mark.unit()
class TestSerdePickleVersion1(TestCase):
    serde = PickleSerde(pickle_version=1)


@pytest.mark.unit()
class TestSerdePickleVersion2(TestCase):
    serde = PickleSerde(pickle_version=2)


@pytest.mark.unit()
class TestSerdePickleVersionHighest(TestCase):
    serde = PickleSerde(
        pickle_version=pickle.HIGHEST_PROTOCOL
    )



# Copyright 2012 Pinterest.com
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A client for falling back to older memcached servers when performing reads.

It is sometimes necessary to deploy memcached on new servers, or with a
different configuration. In these cases, it is undesirable to start up an
empty memcached server and point traffic to it, since the cache will be cold,
and the backing store will have a large increase in traffic.

This class attempts to solve that problem by providing an interface identical
to the Client interface, but which can fall back to older memcached servers
when reads to the primary server fail. The approach for upgrading memcached
servers or configuration then becomes:

 1. Deploy a new host (or fleet) with memcached, possibly with a new
    configuration.
 2. From your application servers, use FallbackClient to write and read from
    the new cluster, and to read from the old cluster when there is a miss in
    the new cluster.
 3. Wait until the new cache is warm enough to support the load.
 4. Switch from FallbackClient to a regular Client library for doing all
    reads and writes to the new cluster.
 5. Take down the old cluster.

Best Practices:
---------------
 - Make sure that the old client has "ignore_exc" set to True, so that it
   treats failures like cache misses. That will allow you to take down the
   old cluster before you switch away from FallbackClient.
"""


class FallbackClient(object):
    def __init__(self, caches):
        assert len(caches) > 0
        self.caches = caches

    def close(self):
        "Close each of the memcached clients"
        for cache in self.caches:
            cache.close()

    def set(self, key, value, expire=0, noreply=True):
        self.caches[0].set(key, value, expire, noreply)

    def add(self, key, value, expire=0, noreply=True):
        self.caches[0].add(key, value, expire, noreply)

    def replace(self, key, value, expire=0, noreply=True):
        self.caches[0].replace(key, value, expire, noreply)

    def append(self, key, value, expire=0, noreply=True):
        self.caches[0].append(key, value, expire, noreply)

    def prepend(self, key, value, expire=0, noreply=True):
        self.caches[0].prepend(key, value, expire, noreply)

    def cas(self, key, value, cas, expire=0, noreply=True):
        self.caches[0].cas(key, value, cas, expire, noreply)

    def get(self, key):
        for cache in self.caches:
            result = cache.get(key)
            if result is not None:
                return result
        return None

    def get_many(self, keys):
        for cache in self.caches:
            result = cache.get_many(keys)
            if result:
                return result
        return []

    def gets(self, key):
        for cache in self.caches:
            result = cache.gets(key)
            if result is not None:
                return result
        return None

    def gets_many(self, keys):
        for cache in self.caches:
            result = cache.gets_many(keys)
            if result:
                return result
        return []

    def delete(self, key, noreply=True):
        self.caches[0].delete(key, noreply)

    def incr(self, key, value, noreply=True):
        self.caches[0].incr(key, value, noreply)

    def decr(self, key, value, noreply=True):
        self.caches[0].decr(key, value, noreply)

    def touch(self, key, expire=0, noreply=True):
        self.caches[0].touch(key, expire, noreply)

    def stats(self):
        # TODO: ??
        pass

    def flush_all(self, delay=0, noreply=True):
        self.caches[0].flush_all(delay, noreply)

    def quit(self):
        # TODO: ??
        pass



# Copyright 2012 Pinterest.com
# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from builtins import bytes as newbytes

from collections import defaultdict
import json
import pytest
import six

from pymemcache.client.base import Client
from pymemcache.exceptions import (
    MemcacheIllegalInputError,
    MemcacheClientError
)
from pymemcache.serde import (
    PickleSerde,
    pickle_serde
)


def get_set_helper(client, key, value, key2, value2):
    result = client.get(key)
    assert result is None

    client.set(key, value, noreply=False)
    result = client.get(key)
    assert result == value

    client.set(key2, value2, noreply=True)
    result = client.get(key2)
    assert result == value2

    result = client.get_many([key, key2])
    assert result == {key: value, key2: value2}

    result = client.get_many([])
    assert result == {}


@pytest.mark.integration()
def test_get_set(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    key = b'key'
    value = b'value'
    key2 = b'key2'
    value2 = b'value2'
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
def test_get_set_newbytes(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    key = newbytes(b'key3')
    value = b'value3'
    key2 = newbytes(b'key4')
    value2 = b'value4'
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
def test_get_set_unicode_key(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module,
                          allow_unicode_keys=True)
    client.flush_all()

    key = u"こんにちは"
    value = b'hello'
    key2 = 'my☃'
    value2 = b'value2'
    get_set_helper(client, key, value, key2, value2)


@pytest.mark.integration()
def test_add_replace(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.add(b'key', b'value', noreply=False)
    assert result is True
    result = client.get(b'key')
    assert result == b'value'

    result = client.add(b'key', b'value2', noreply=False)
    assert result is False
    result = client.get(b'key')
    assert result == b'value'

    result = client.replace(b'key1', b'value1', noreply=False)
    assert result is False
    result = client.get(b'key1')
    assert result is None

    result = client.replace(b'key', b'value2', noreply=False)
    assert result is True
    result = client.get(b'key')
    assert result == b'value2'


@pytest.mark.integration()
def test_append_prepend(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.append(b'key', b'value', noreply=False)
    assert result is False
    result = client.get(b'key')
    assert result is None

    result = client.set(b'key', b'value', noreply=False)
    assert result is True
    result = client.append(b'key', b'after', noreply=False)
    assert result is True
    result = client.get(b'key')
    assert result == b'valueafter'

    result = client.prepend(b'key1', b'value', noreply=False)
    assert result is False
    result = client.get(b'key1')
    assert result is None

    result = client.prepend(b'key', b'before', noreply=False)
    assert result is True
    result = client.get(b'key')
    assert result == b'beforevalueafter'


@pytest.mark.integration()
def test_cas(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()
    result = client.cas(b'key', b'value', b'1', noreply=False)
    assert result is None

    result = client.set(b'key', b'value', noreply=False)
    assert result is True

    # binary, string, and raw int all match -- should all be encoded as b'1'
    result = client.cas(b'key', b'value', b'1', noreply=False)
    assert result is False
    result = client.cas(b'key', b'value', '1', noreply=False)
    assert result is False
    result = client.cas(b'key', b'value', 1, noreply=False)
    assert result is False

    result, cas = client.gets(b'key')
    assert result == b'value'

    result = client.cas(b'key', b'value1', cas, noreply=False)
    assert result is True

    result = client.cas(b'key', b'value2', cas, noreply=False)
    assert result is False


@pytest.mark.integration()
def test_gets(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.gets(b'key')
    assert result == (None, None)

    result = client.set(b'key', b'value', noreply=False)
    assert result is True
    result = client.gets(b'key')
    assert result[0] == b'value'


@pytest.mark.integration()
def test_delete(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.delete(b'key', noreply=False)
    assert result is False

    result = client.get(b'key')
    assert result is None
    result = client.set(b'key', b'value', noreply=False)
    assert result is True
    result = client.delete(b'key', noreply=False)
    assert result is True
    result = client.get(b'key')
    assert result is None


@pytest.mark.integration()
def test_incr_decr(client_class, host, port, socket_module):
    client = Client((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.incr(b'key', 1, noreply=False)
    assert result is None

    result = client.set(b'key', b'0', noreply=False)
    assert result is True
    result = client.incr(b'key', 1, noreply=False)
    assert result == 1

    def _bad_int():
        client.incr(b'key', b'foobar')

    with pytest.raises(MemcacheClientError):
        _bad_int()

    result = client.decr(b'key1', 1, noreply=False)
    assert result is None

    result = client.decr(b'key', 1, noreply=False)
    assert result == 0
    result = client.get(b'key')
    assert result == b'0'


@pytest.mark.integration()
def test_touch(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    result = client.touch(b'key', noreply=False)
    assert result is False

    result = client.set(b'key', b'0', 1, noreply=False)
    assert result is True

    result = client.touch(b'key', noreply=False)
    assert result is True

    result = client.touch(b'key', 1, noreply=False)
    assert result is True


@pytest.mark.integration()
def test_misc(client_class, host, port, socket_module):
    client = Client((host, port), socket_module=socket_module)
    client.flush_all()


@pytest.mark.integration()
def test_serialization_deserialization(host, port, socket_module):
    class JsonSerde(object):
        def serialize(self, key, value):
            return json.dumps(value).encode('ascii'), 1

        def deserialize(self, key, value, flags):
            if flags == 1:
                return json.loads(value.decode('ascii'))
            return value

    client = Client((host, port), serde=JsonSerde(),
                    socket_module=socket_module)
    client.flush_all()

    value = {'a': 'b', 'c': ['d']}
    client.set(b'key', value)
    result = client.get(b'key')
    assert result == value


def serde_serialization_helper(client_class, host, port,
                               socket_module, serde):
    def check(value):
        client.set(b'key', value, noreply=False)
        result = client.get(b'key')
        assert result == value
        assert type(result) is type(value)

    client = client_class((host, port), serde=serde,
                          socket_module=socket_module)
    client.flush_all()

    check(b'byte string')
    check(u'unicode string')
    check('olé')
    check(u'olé')
    check(1)
    check(123123123123123123123)
    check({'a': 'pickle'})
    check([u'one pickle', u'two pickle'])
    testdict = defaultdict(int)
    testdict[u'one pickle']
    testdict[b'two pickle']
    check(testdict)


@pytest.mark.integration()
def test_serde_serialization(client_class, host, port, socket_module):
    serde_serialization_helper(client_class, host, port,
                               socket_module, pickle_serde)


@pytest.mark.integration()
def test_serde_serialization0(client_class, host, port, socket_module):
    serde_serialization_helper(
        client_class, host, port,
        socket_module,
        PickleSerde(pickle_version=0))


@pytest.mark.integration()
def test_serde_serialization2(client_class, host, port, socket_module):
    serde_serialization_helper(
        client_class, host, port,
        socket_module,
        PickleSerde(pickle_version=2))


@pytest.mark.integration()
def test_errors(client_class, host, port, socket_module):
    client = client_class((host, port), socket_module=socket_module)
    client.flush_all()

    def _key_with_ws():
        client.set(b'key with spaces', b'value', noreply=False)

    with pytest.raises(MemcacheIllegalInputError):
        _key_with_ws()

    def _key_with_illegal_carriage_return():
        client.set(b'\r\nflush_all', b'value', noreply=False)

    with pytest.raises(MemcacheIllegalInputError):
        _key_with_illegal_carriage_return()

    def _key_too_long():
        client.set(b'x' * 1024, b'value', noreply=False)

    with pytest.raises(MemcacheClientError):
        _key_too_long()

    def _unicode_key_in_set():
        client.set(six.u('\u0FFF'), b'value', noreply=False)

    with pytest.raises(MemcacheClientError):
        _unicode_key_in_set()

    def _unicode_key_in_get():
        client.get(six.u('\u0FFF'))

    with pytest.raises(MemcacheClientError):
        _unicode_key_in_get()

    def _unicode_value_in_set():
        client.set(b'key', six.u('\u0FFF'), noreply=False)

    with pytest.raises(MemcacheClientError):
        _unicode_value_in_set()


@pytest.mark.integration()
def test_tls(client_class, tls_host, tls_port, socket_module, tls_context):
    client = client_class(
        (tls_host, tls_port),
        socket_module=socket_module,
        tls_context=tls_context
    )
    client.flush_all()

    key = b'key'
    value = b'value'
    key2 = b'key2'
    value2 = b'value2'
    get_set_helper(client, key, value, key2, value2)
